#!/usr/bin/python3

import sys
import math

def mmm(cpu,key):
    has_pre = False
    if cpu.prefix != None:
        has_pre = True
        pre = cpu.prefix
        cpu.prefix = None

    if key == 0b000:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['edx']+cpu.reg['esi']
        else:
            return cpu.reg['ds']+cpu.reg['edx']+cpu.reg['esi']
    if key == 0b001:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['ebx']+cpu.reg['edi']
        else:
            return cpu.reg['ds']+cpu.reg['ebx']+cpu.reg['edi']
    if key == 0b010:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['ebp']+cpu.reg['esi']
        else:
            return cpu.reg['ss']+cpu.reg['ebp']+cpu.reg['esi']
    if key == 0b011:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['ebp']+cpu.reg['edi']
        else:
            return cpu.reg['ss']+cpu.reg['ebp']+cpu.reg['edi']
    if key == 0b100:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['esi']
        else:
            return cpu.reg['ds']+cpu.reg['esi']
    if key == 0b101:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['edi']
        else:
            return cpu.reg['ds']+cpu.reg['edi']
    if key == 0b110:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['ebp']
        else:
            return cpu.reg['ss']+cpu.reg['ebp']
    if key == 0b111:
        if has_pre:
            return cpu.reg[pre]+cpu.reg['ebx']
        else:
            return cpu.reg['ds']+cpu.reg['ebx']



rrr = {
    0b000 : 'eax',
    0b001 : 'ecx',
    0b010 : 'edx',
    0b011 : 'ebx',
    0b101 : 'ebp',
    0b110 : 'esi',
    0b111 : 'edi',
}

sss = {
    0b000 : 'es',
    0b001 : 'cs',
    0b010 : 'ss',
    0b011 : 'ds',
}

def op_halt(cpu):
    cpu.stop = True


def get_bit_range(byte, start, len=None, bit_length = None):
    if bit_length == None:
        bit_length = byte.bit_length()
    if len == None:
        len = bit_length -start+1
    tmp = (1<< (bit_length -start+1)) -1
    #print(bit_length,bin(tmp))
    val = (byte & tmp) >> (bit_length -start-len+1)
    #print(bin(val))
    return val

    
def op_jmp(cpu):
    iop=int.from_bytes(cpu.current_machine_code, 'big')
    cpu.reg['eip']=cpu.reg[rrr[get_bit_range(iop, 14)]]

def op_loop(cpu):
    iop=int.from_bytes(cpu.current_machine_code[1:], 'big',signed=True)
    if cpu.reg['ecx'] !=0:
        cpu.reg['eip']+=iop
    cpu.reg['ecx']-=1
    



def op_movi(cpu):
    iop=int.from_bytes(cpu.current_machine_code, 'big')
    Reg = rrr[get_bit_range(iop, 6,3)]
    Imm = int.from_bytes(cpu.current_machine_code[1:], 'little')
    cpu.reg[Reg]=Imm

def op_movmi(cpu):
    iop=int.from_bytes(cpu.current_machine_code, 'big')
    Imm = int.from_bytes(cpu.current_machine_code[2:], 'little')
    cpu.mem[mmm(cpu,get_bit_range(iop, 14,3))-0x7c00]=Imm

def op_movsr(cpu):
    iop=int.from_bytes(cpu.current_machine_code, 'big')
    Seg = sss[get_bit_range(iop, 11,3)]
    Reg16 = cpu.reg[rrr[get_bit_range(iop, 14,3)]]
    cpu.reg[Seg]=Reg16

def op_movrr_rm(cpu):
    iop=int.from_bytes(cpu.current_machine_code, 'big')
    #print(bin(iop))
    if get_bit_range(iop, 9,2) == 0b11:
        #print(Reg1)
        Reg1 = rrr[get_bit_range(iop, 11,3)]
        Reg2 = cpu.reg[rrr[get_bit_range(iop, 14,3)]]
        cpu.reg[Reg1]=Reg2
    else:
        Reg = rrr[get_bit_range(iop, 11,3)]
        Mem = cpu.getmem(mmm(cpu,get_bit_range(iop, 14,3)),1)
        #print(Reg)
        cpu.reg[Reg] = int.from_bytes(Mem, 'little')
        #cpu.mem[cpu.reg[Reg]-0x7c00]=int.from_bytes(Mem, 'little')




def op_movmr(cpu):
    iop=int.from_bytes(cpu.current_machine_code, 'big')
    Reg = cpu.reg[rrr[get_bit_range(iop, 11,3)]]
    Mem = mmm(cpu,get_bit_range(iop, 14,3))
    #print('es:di:',hex(Mem))
    cpu.mem[Mem-0x7c00] = Reg
    

def op_inc(cpu):
    iop=int.from_bytes(cpu.current_machine_code, 'big')
    #print(rrr[get_bit_range(iop, 6,3,8)])
    #print(cpu.reg[rrr[get_bit_range(iop, 6,3,8)]])
    cpu.reg[rrr[get_bit_range(iop, 6,3,8)]] +=1
    #print(cpu.reg[rrr[get_bit_range(iop, 6,3,8)]])

def op_cs(cpu):
    cpu.prefix = 'cs'
def op_ds(cpu):
    cpu.prefix = 'ds'
def op_es(cpu):
    cpu.prefix = 'es'
def op_fs(cpu):
    cpu.prefix = 'fs'
def op_gs(cpu):
    cpu.prefix = 'gs'
def op_ss(cpu):
    cpu.prefix = 'ss'


oplist =[
    {
        "key":bytearray([0b10011010]),
        "oplen":8,
        "len":1,
        "name":"CALL Far",
        "function":None,
    },
    {
        "key":bytearray([0b11111111]),
        "oplen":8,
        "len":2,
        "name":"jmp",
        "function":op_jmp,
    },
    {
        "key":bytearray([0b01000]),
        "oplen":5,
        "len":1,
        "name":"inc",
        "function":op_inc,
    },
    {
        "key":bytearray([0b11100010]),
        "oplen":8,
        "len":2,
        "name":"loop",
        "function":op_loop,
    },
    {
        "key":bytearray([0b11110100]),
        "oplen":8,
        "len":1,
        "name":"hlt",
        "function":op_halt,
    },
    {
        "key":bytearray([0b1011]),
        "oplen":4,
        "len":3,
        "name":"MOV Reg,Imm",
        "function":op_movi,
    },
    {
        "key":bytearray([0b1000101]),
        "oplen":7,
        "len":2,
        "name":"MOV Reg,Reg or MOV Reg,Mem",
        "function":op_movrr_rm,
    },
    {
        "key":bytearray([0b10001110]),
        "oplen":8,
        "len":2,
        "name":"MOV Seg,Reg16",
        "function":op_movsr,
    },
    {
        "key":bytearray([0b1000100]),
        "oplen":7,
        "len":2,
        "name":"MOV Mem,Reg",
        "function":op_movmr,
    },
    {
        "key":bytearray([0b1100011]),
        "oplen":7,
        "len":3,
        "name":"MOV Mem,Imm",
        "function":op_movmi,
    },
    {
        "key":bytearray([0b00101110]),
        "oplen":8,
        "len":1,
        "name":"CS:",
        "function":op_cs,
    },
    {
        "key":bytearray([0b00111110]),
        "oplen":8,
        "len":1,
        "name":"DS:",
        "function":op_ds,
    },
    {
        "key":bytearray([0b00100110]),
        "oplen":8,
        "len":1,
        "name":"ES:",
        "function":op_es,
    },
    {
        "key":bytearray([0b01100100]),
        "oplen":8,
        "len":1,
        "name":"FS:",
        "function":op_fs,
    },
    {
        "key":bytearray([0b01100101]),
        "oplen":8,
        "len":1,
        "name":"GS:",
        "function":op_gs,
    },
    {
        "key":bytearray([0b00110110]),
        "oplen":8,
        "len":1,
        "name":"SS:",
        "function":op_ss,
    }
]


class cpu:
    def __init__(self):
        self.prefix = None
        self.stop = False
        self.reg={
            'eip':0x7c00,
            'eax':0xaa55,
            'ebx':0,
            'ecx':0,
            'edx':0x80,
            'ebp':0,
            'esi':0,
            'edi':0xffac,
            'cs':0,
            'ds':0,
            'es':0,
            'ss':0,
        }
        self.mem = bytearray(' '*50000, encoding='utf8')
        self.mem[0xb800-0x7c00+10*80*2+60:0xb800-0x7c00+10*80*2+60+len('hello world!')*2] = bytearray('h e l l o   w o r l d ! '.encode())

    def getmem(self,start,len):
        return self.mem[start-0x7c00:start-0x7c00+len]

    def load_program(self, path):
        f = open(path, mode='rb')
        code = f.read()
        self.mem[0:len(code)]=bytearray(code)
        f.close()


    def print_reg(self):
        for k, v in self.reg.items():
            print(k,':', hex(v))


    def run(self,step):
        i = 0
        while not self.stop:
            i+=1
            if i>step:
                break;

            for x in oplist:
                bytelen=math.ceil(x['oplen']/8)
                byteop = int.from_bytes(self.getmem(self.reg['eip'],bytelen), byteorder='big')>>(8-x['oplen'])
                opkey = int.from_bytes(x['key'], byteorder='big')
                #print(bin(byteop),bin(opkey))
                if byteop == opkey:
                    current_op = x
                    machine_code = self.getmem(self.reg['eip'],x['len'])
                    break;
            else:
                print(hex(self.reg['eip']),'代码中含有未实现指令:');
                print(bin(int.from_bytes(self.getmem(self.reg['eip'],2), byteorder='big')))
                exit();


            tmp = int.from_bytes(machine_code, byteorder='big')
            print(i,'eip:',hex(self.reg['eip']),hex(tmp),bin(tmp),current_op['name'])
            #print(self.mem.decode('ascii'))
            self.print_reg()
            self.reg['eip']+=current_op['len']
            self.current_op = current_op
            self.current_machine_code = machine_code
            current_op['function'](self)


if len(sys.argv) != 2:
    print("usage: %s <images>"%(sys.argv[0]))
    eixt()


c=cpu()
c.load_program(sys.argv[1])
c.run(100)
import json
import asyncio
import websockets

async def hello(websocket, path):
    c=cpu()
    c.load_program(sys.argv[1])
    c.run(100)
    data = {
        "screen" : list(c.mem[0xb800-0x7c00:0xb800-0x7c00+80*25*2]),
        "reg" : c.reg,
    }
    await websocket.send(json.dumps(data))

start_server = websockets.serve(hello, "localhost", 6999)

asyncio.get_event_loop().run_until_complete(start_server)
asyncio.get_event_loop().run_forever()

